/*
 *   Copyright (c) 2007 Paolo Capriotti <p.capriotti@gmail.com>
 *   Copyright (c) 2009 Aaron Seigo <aseigo@kde.org>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License as
 *   published by the Free Software Foundation; either version 2, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "plasma/private/wallpaperrenderthread_p.h"

#include <QPainter>
#include <QFile>
#include <QSvgRenderer>

#include <kdebug.h>

namespace Plasma
{

WallpaperRenderThread::WallpaperRenderThread(QObject *parent)
    : QThread(parent),
      m_lastToken(-1)
{
    m_abort = false;
}

WallpaperRenderThread::~WallpaperRenderThread()
{
    {
        // abort computation
        QMutexLocker lock(&m_mutex);
        m_abort = true;
    }

    wait();
}

int WallpaperRenderThread::render(const QString &file,
                                  const QSize &size,
                                  Wallpaper::ResizeMethod method,
                                  const QColor &color)
{
    int token;
    RenderTask task;
    task.file = file;
    task.color = color;
    task.resizeMethod = method;
    task.size = size;
    {
        QMutexLocker lock(&m_mutex);
        task.token = ++m_lastToken;
        m_tasks.enqueue(task);
    }

    if (!isRunning()) {
        start();
    }

    return token;
}

void WallpaperRenderThread::run()
{
    qreal ratio;
    RenderTask task;

    forever {
        {
            QMutexLocker lock(&m_mutex);

            if (m_tasks.isEmpty() || m_abort) {
                return;
            }

            // load all parameters in nonshared variables
            task = m_tasks.dequeue();
            ratio = task.size.width() / qreal(task.size.height());
        }

        QImage result(task.size, QImage::Format_ARGB32_Premultiplied);
        result.fill(task.color.rgba());

        if (task.file.isEmpty() || !QFile::exists(task.file)) {
            emit done(task.token, result, task.file, task.size, task.resizeMethod, task.color);
            break;
        }

        QPoint pos(0, 0);
        bool tiled = false;
        bool scalable = task.file.endsWith("svg") || task.file.endsWith("svgz");
        QSize scaledSize;
        QImage img;

        // set image size
        QSize imgSize;
        if (scalable) {
            // scalable: image can be of any size
            imgSize = task.size;
        } else {
            // otherwise, use the natural size of the loaded image
            img = QImage(task.file);
            imgSize = img.size();
            //kDebug() << "loaded with" << imgSize << ratio;
        }

        // if any of them is zero we may run into a div-by-zero below.
        if (imgSize.width() < 1) {
            imgSize.setWidth(1);
        }

        if (imgSize.height() < 1) {
            imgSize.setHeight(1);
        }

        if (ratio < 1) {
            ratio = 1;
        }

        // set render parameters according to resize mode
        switch (task.resizeMethod)
        {
        case Wallpaper::ScaledResize:
            imgSize *= ratio;
            scaledSize = task.size;
            break;
        case Wallpaper::CenteredResize:
            scaledSize = imgSize;
            pos = QPoint((task.size.width() - scaledSize.width()) / 2,
                        (task.size.height() - scaledSize.height()) / 2);

            //If the picture is bigger than the screen, shrink it
            if (task.size.width() < imgSize.width() && imgSize.width() > imgSize.height()) {
                int width = task.size.width();
                int height = width * scaledSize.height() / imgSize.width();
                scaledSize = QSize(width, height);
                pos = QPoint((task.size.width() - scaledSize.width()) / 2,
                             (task.size.height() - scaledSize.height()) / 2);
            } else if (task.size.height() < imgSize.height()) {
                int height = task.size.height();
                int width = height * imgSize.width() / imgSize.height();
                scaledSize = QSize(width, height);
                pos = QPoint((task.size.width() - scaledSize.width()) / 2,
                             (task.size.height() - scaledSize.height()) / 2);
            }

            break;
        case Wallpaper::MaxpectResize: {
            imgSize *= ratio;
            float xratio = (float) task.size.width() / imgSize.width();
            float yratio = (float) task.size.height() / imgSize.height();
            if (xratio > yratio) {
                int height = task.size.height();
                int width = height * imgSize.width() / imgSize.height();
                scaledSize = QSize(width, height);
            } else {
                int width = task.size.width();
                int height = width * imgSize.height() / imgSize.width();
                scaledSize = QSize(width, height);
            }
            pos = QPoint((task.size.width() - scaledSize.width()) / 2,
                        (task.size.height() - scaledSize.height()) / 2);
            break;
        }
        case Wallpaper::ScaledAndCroppedResize: {
            imgSize *= ratio;
            float xratio = (float) task.size.width() / imgSize.width();
            float yratio = (float) task.size.height() / imgSize.height();
            if (xratio > yratio) {
                int width = task.size.width();
                int height = width * imgSize.height() / imgSize.width();
                scaledSize = QSize(width, height);
            } else {
                int height = task.size.height();
                int width = height * imgSize.width() / imgSize.height();
                scaledSize = QSize(width, height);
            }
            pos = QPoint((task.size.width() - scaledSize.width()) / 2,
                        (task.size.height() - scaledSize.height()) / 2);
            break;
        }
        case Wallpaper::TiledResize:
            scaledSize = imgSize;
            tiled = true;
            break;
        case Wallpaper::CenterTiledResize:
            scaledSize = imgSize;
            pos = QPoint(
                -scaledSize.width() +
                    ((task.size.width() - scaledSize.width()) / 2) % scaledSize.width(),
                -scaledSize.height() +
                    ((task.size.height() - scaledSize.height()) / 2) % scaledSize.height());
            tiled = true;
            break;
        }

        QPainter p(&result);
        //kDebug() << token << scalable << scaledSize << imgSize;
        if (scalable) {
            // tiling is ignored for scalable wallpapers
            QSvgRenderer svg(task.file);
            if (m_abort) {
                return;
            }
            svg.render(&p);
        } else {
            if (scaledSize != imgSize) {
                img = img.scaled(scaledSize, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
            }

            if (m_abort) {
                return;
            }

            if (tiled) {
                for (int x = pos.x(); x < task.size.width(); x += scaledSize.width()) {
                    for (int y = pos.y(); y < task.size.height(); y += scaledSize.height()) {
                        p.drawImage(QPoint(x, y), img);
                        if (m_abort) {
                            return;
                        }
                    }
                }
            } else {
                p.drawImage(pos, img);
            }
        }

        // signal we're done
        emit done(task.token, result, task.file, task.size, task.resizeMethod, task.color);
    }
}

} // namespace Plasma

#include "wallpaperrenderthread_p.moc"

